#!/usr/bin/env python
"""
###############################################################################
# Program : Climate Compatible Growth
# Author  : Miriam Zachau Walker, Flora Charbonnier
# Date    : March 2021
# Purpose : Create csv with values for base SAND file from data collection file
##############################################################################
Instructions

Step 0 - Install relevant packages (if necessary)
    1. Open the terminal
    2. Navigate to the relevant directory
        cd /path/to/CCG_data_to_SAND.py
    3. Activate the relevant virtual environment (optional / if applicable)
        python3 -m venv venv
        source venv/bin/activate
        python3 -m pip install --upgrade pip
    4. Download the package dependencies
        python3 -m pip install -r requirements.txt

For further help (or if you don't already have pip), please check:
https://packaging.python.org/tutorials/installing-packages/

Step 1 - Update user inputs
The user should update:
    1. PATH_INPUT:        path to folder with data collection files
    2. PATH_SAND:         path to folder with empty SAND file
    3. PATH_COUNTRIES:    path to folder with countries file
    4. PATH_SAND_OUT:     path to folder where save csv with values for
                            Base SAND (default same as SAND)
    5. countries:         list of countries for which SAND files should be created

Step 2 - Run this script (twice if needed, if the adequate BASE sand file is not available, a link to the download page will be printed on screen)
    1. Open the terminal
    2. Activate the relevant virtual environment (optional / if applicable)
        cd /path/to/CCG_data_to_SAND.py
        source venv/bin/activate
    3. Type python THIS_FILE.py , then press ENTER
        python3 CCG_data_to_SAND_40TS_v2.py
        or
        python CCG_data_to_SAND_40TS_v2.py
        Note: if you are not in the right directory already, THIS_FILE should include the file path,
        e.g. python /Users/username/path_folders/CCG - Starter Kit/clickSAND/CCG_data_to_SAND.py

Step 3 - Copy values into parameter sheet
    1. Create a new empty base SAND Excel file by copying ‘Base SAND.xlsm’ and renaming it to “New_{country_name}_Base_SAND.xlsm"
    2. Ensure all rows are visible and no filtering is active in country Base SAND.xlsm
    3. Open the “{country_name} Base SAND parameters.csv" file that was generated by the Python script
    4. Copy the contents from the entire .csv file
    5. Paste values only into the Parameters tab of the “New_{country_name}_Base_SAND.xlsm” file.
    Note: This might only work in desktop Excel version, not web app


Note - For regions outside of Sub Saharan Africa
Update blank SAND filenames for other regions and uncomment those sections as needed

"""
#%% ######################### Import Required Modules ##########################
import time
import os
import traceback
import numpy as np
import pandas as pd
import shutil

#%% ############################# USER INPUTS ################################
# update list of countries to create base SAND files for
COUNTRIES = ['SUDAN']

# update file paths
PATH_INPUT     = 'path/to/directory/Data in Brief/'
PATH_COUNTRIES = PATH_INPUT
PATH_SAND      = PATH_INPUT
PATH_SAND_OUT  = PATH_INPUT

assert len(PATH_INPUT) > 1, \
    "update PATH_INPUT value (folder with 'New country_name Data Collection.xlsx')"
assert len(PATH_COUNTRIES) > 1, \
    "update PATH_COUNTRIES value (folder with 'Starter Kit - List of Countries and Regions.xlsx')"
assert len(PATH_SAND) > 1, \
    "update PATH_SAND value (folder with empty SAND file)"
assert len(PATH_SAND_OUT) > 1, \
    "update PATH_SAND_OUT value (folder where to save output Base SAND (default same as SAND)"
assert len(COUNTRIES) > 0, \
    "update COUNTRIES list of countries to create base SAND files for"

#%%########################## Define functions ########################
def check_offshore_wind(cap_inv):
    """ check status if offshore wind is possible for a country
    first check if any potential listed(>0) and if so, then True
    if none, check if zeros have been filled in or spaces are empty
          by summing NaN values"""

    # obtain capacity constraints for offshore wind
    offshore = cap_inv.loc[
        (cap_inv['Capacity Constraint'] == 'Total Annual Maximum Capacity')
        & (cap_inv['Code'] == 'PWRWND002'), years]

    if np.nansum(offshore) > 0:   # if some capacities are larger than zero
        offshore_possible  = True # offshore is possible
    else:
        sum_nans = offshore.isna().sum(axis = 1).values[0]
        # if all constraints cells are empty, no contraints added, so
        # offshore wind is possible, otherwise else not
        offshore_possible = sum_nans >= len(years)

    return offshore_possible

def condition_update(param, input_data):
    """ boolean to determine whether to update a row
        first checks if parameter is resid cap
            if not, then returns true and will update always
        if res cap, then checks if that tech is installed
            to determine whether to update
            True = installed
            False = not installed
    """
    conditional_params = ['ResidualCapacity']
    if param not in conditional_params:
        # no condition, always update params using the inputdata
        return True
    else:
        if param == 'ResidualCapacity':
        # update parameter sheet ONLY IF there is resid cap for that tech
            return np.nansum(input_data) > 0

#%% ################ Define general info for all countries #############
years            = range(2015, 2071)
years_str        = [str(x) for x in years]
years_past       = [str(x) for x in range(2015, 2021)]

# list countries
filename_countries = 'Starter Kit - List of Countries and Regions.xlsx'

#%% ######################### Loop over countries ##########################
for country in COUNTRIES:
    try:
        t_start = time.time() # record the running time
        print(f'starting new base SAND file for {country}')

        # Define filenames
        # input data file
        filename_data = 'New ' + str(country) + ' Data Collection.xlsx'
        if not os.path.exists(os.path.join(PATH_INPUT, filename_data)):
            shutil.copy(os.path.join(PATH_INPUT, "Base Data Collection File.xlsx"),
                        os.path.join(PATH_INPUT, filename_data))

        # output name of new file
        new_file = country + ' Base SAND.xlsm'
        #%% ######################### Load input data ##########################
        # name of sheets in input data file
        sheets = ['3.3 Output Activity Ratios', '3.6 Output Capacity Factors',
               '3.7 ResCap data (auto)', '3.8 Capacity & Inv Constraints',
               '4.1 Accumulated Annual Demand', '4.2 Specified Annual Demand',
               '4.2Specified Dem Profile Output','5.1 Activity']

        # Row (0-indexed) to use for the column labels
        headers = [1, 1, 1, 1, 2, 2, 2, 1]

        # Column (0-indexed) to use as the row labels of the DataFrame
        index_cols = [None, 0, None, None, None, None, None, 0, None]

        # label of data in data dictionary
        # chosen to correspond to parameters, where possible
        label = ['OutputActivityRatio', 'CapacityFactor', 'ResidualCapacity',
             'TotalAnnualMaxCapacity', 'AccumulatedAnnualDemand',
             'SpecifiedAnnualDemand', 'SpecifiedDemandProfile', 'Activity']

        # create dictionary of input data DataFrames
        # read each sheet as dataframe, store in dictionary of dataframes
        data = {}
        for i in range(len(sheets)):
            data[label[i]] = pd.read_excel(os.path.join(PATH_INPUT, filename_data),
                            sheet_name = sheets[i], header = headers[i],
                            index_col = index_cols[i])

        # fix years for ResidualCapacity
        # because 3.7 ResCap data (auto) only goes to 2050
        list_entries       = list(data['ResidualCapacity'].columns)
        list_entries[-20:] = range(2051, 2071)
        data['ResidualCapacity'].columns = list_entries

        #%%######################### Create output spreadsheet #####################
        # choose which base file to use depending on region/country
        # find which region the country is in in the list of countries
        list_countries = pd.read_excel(PATH_COUNTRIES + filename_countries)
        region = list_countries['Region '].loc[list_countries['Name of the Country']
                                           == country.upper()].values[0]

        print(f"region: {region}")

        if region == 'Sub-Saharan Africa':
            # specify which countries are in which groups
            NGS_countries = ['Nigeria', 'South Africa']
            NGS_COA_countries = []

            # specify which base file to use for each set of countries
            if country in NGS_countries:
                filename_empty_SAND  = 'Africa_NGS_base_SAND.xlsm'
                url = 'https://zenodo.org/record/6036135'
            elif country in NGS_COA_countries:
                filename_empty_SAND  = 'Africa_COA_NGS_base_SAND.xlsm'
                url = 'https://zenodo.org/record/6033764'
            else:
            # use this as default unless in other group
                filename_empty_SAND  = 'Africa_base_SAND.xlsm'
                url = 'https://zenodo.org/record/6011229'

        elif region == 'South America':
            filename_empty_SAND  = 'South_America_COA_NGS_base_SAND.xlsm'
            url = 'https://zenodo.org/record/6036192'
        elif region == 'Developing Asia':
            filename_empty_SAND  = 'Asia_COA_NGS_base_SAND.xlsm'
            url = 'https://zenodo.org/record/6036147'
        else:
            print('ERROR: Region not found')
            break

        # check whether the base SAND file required has been downloaded
        path_empty_SAND = os.path.join(PATH_SAND, filename_empty_SAND)
        if not os.path.exists(path_empty_SAND):
            print(f"Please copy {filename_empty_SAND} from {url} into the directory {PATH_SAND}")
        else:
            ####### create blank parameters spreadsheet based on empty SAND doc
            # base file for output data
            df_parameters = pd.read_excel(path_empty_SAND,
                                      sheet_name = 'Parameters')
            #%%######################### Check if offshore wind possible #########################
            offshore_wind_possible = check_offshore_wind(data['TotalAnnualMaxCapacity'])

            #%%######### update parameters where data is available for all years

            # update : list of dictionaries of information to pass from input data to df_parameters, corresponds to rows
            # Parameter : the parameter to be updated in df_parameters
            # Parameter_col : if specified, the 'Parameter_col' column in input_data needs to be equal to the Parameter
            # TECHNOLOGY : if specified, the TECHNOLOGY to be updated in df_parameters
            # TECHNOLOGY_col : if TECHNOLOGY is specified, select TECHNOLOGY rows of 'TECHNOLOGY_col' columns in input_data
            # FUEL : if specified, the FUEL to be updated in df_parameters
            # FUEL_col : if FUEL is specified, select rows equal to the FUEL in 'FUEL_col' columns in input_data
            # input_param : if the column name corresponding to the Parameter is not the same in input_data as in df_parameter, specify
            # data_entry : if the data entry does not correspond to the Parameter (e.g. if multiple parameters use the same data entry), specify

            # initialise list of update instructions
            # these correspond to which rows to update on the Parameters tab
            updates = []

            # OutputActivityRatio
            techs = ['PWRTRN', 'PWRDIST']
            fuels = ['ELC002', 'ELC003']
            for tech, fuel in zip(techs, fuels):
                updates += [{'Parameter'      : 'OutputActivityRatio',
                        'TECHNOLOGY'     : tech,
                        'FUEL'           : fuel,
                        'TECHNOLOGY_col' : 'Technology Code',
                        'FUEL_col'       : 'Output Fuel (Code)'}]

            # Residual Capacity
            pwr_techs_rescap = [
                'PWRBIO001',
                'PWRCOA001',
                'PWRCSP001', 'PWRCSP002',
                'PWRDIST',
                'PWRGEO',
                 'PWRHYD001', 'PWRHYD002', 'PWRHYD003', 'PWRHYD004',
                'PWRNGS001', 'PWRNGS002',
                'PWRNUC',
                'PWROHC001', 'PWROHC002', 'PWROHC003',
                'PWRSOL001', 'PWRSOL001S', 'PWRSOL002',
                'PWRTRN',
                'PWRWND001','PWRWND001S', 'PWRWND002',
                'UPSREF001', 'UPSREF002']

            for tech in pwr_techs_rescap:
                updates += [{'Parameter'      : 'ResidualCapacity',
                        'TECHNOLOGY'     : tech,
                        'TECHNOLOGY_col' : 'Group',
                        'Parameter_col'  : 'Power Plant'}]
            # total annual max capacity
            # only include offshore wind constraints if impossible
            techs = ['PWRHYD001', 'PWRHYD002', 'PWRHYD003', 'PWRHYD004', 'PWRGEO']
            if not offshore_wind_possible:
                techs += ['PWRWND002']
            for tech in techs:
                updates += [{'Parameter'  : 'TotalAnnualMaxCapacity',
                        'input_param'    : 'Total Annual Maximum Capacity',
                        'TECHNOLOGY'     : tech,
                        'TECHNOLOGY_col' : 'Code',
                        'Parameter_col'  : 'Capacity Constraint'}]
            # update accumulated annual demand
            fuels = ['TRAMCY', 'TRACAR', 'TRABUS', 'INDHEH', 'INDHEL',
                 'RESCKN', 'COMHEL', 'RESHEL']
            for fuel in fuels:
                updates += [{'Parameter' : 'AccumulatedAnnualDemand',
                        'FUEL'      : fuel,
                        'FUEL_col'  : 'Commodity Code'}]
            # Specified Annual Demand
            fuels_spec_dem = ['INDELC', 'RESELC', 'COMELC']
            # this list will be used again later
            for fuel in fuels_spec_dem:
                updates += [{'Parameter' : 'SpecifiedAnnualDemand',
                        'FUEL'      : fuel,
                        'FUEL_col'  : 'Commodity Code'}]
            # total technology annual activity upper and lower limit
            techs      = ['PWRTRNIMP', 'PWRTRNEXP']
            parameters = ['TotalTechnologyAnnualActivityLowerLimit',
                      'TotalTechnologyAnnualActivityUpperLimit']
            for tech in techs:
                for param in parameters:
                    updates += [{'Parameter'      : param,
                            'TECHNOLOGY'     : tech,
                            'TECHNOLOGY_col' : 'Technology Code',
                            'Parameter_col'  : 'Constraint',
                            'data_entry'      : 'Activity'}]

            ######## loop over updates rules dictionaries to update each row
            no_data, too_much_data = [], []
            for update in updates:
            # for i in range(len(updates)):
                # initialise dictionaries to store column names,
                # input row selection booleans and output row selection booleans
                column, input_bools, output_bools = [{} for _ in range(3)]
                param = update['Parameter']

                # the label for the data entry data_entry is the same as the parameter
                # unless data_entry is specified
                data_entry = update['data_entry'] if 'data_entry' in update else param

                # the name input_param of the parameter to be selected in the
                # data is the same as the parameter unless 'input_param' is specified
                input_param = update['input_param'] if 'input_param' in update \
                    else update['Parameter']
                # loop through TECHNOLOGY and FUEL to obtain the booleans for selecting
                # over data and df_parameters
                for e in ['TECHNOLOGY', 'FUEL']:
                    if e not in update:
                        # if no TECHNOLOGY/FUEL was specified, all rows are valid
                        input_bools[e]  = [True for _ in range(len(data[data_entry]))]
                        output_bools[e] = [True for _ in range(len(df_parameters[e]))]
                    else:
                        # else, compute boolean for each row specifyng whether values
                        # in column correponding to TECH/FUEL are the specified TECH/FUEL
                        column[e]       = update[e + '_col']
                        input_bools[e]  = (data[data_entry][column[e]] == update[e])
                        output_bools[e] = (df_parameters[e] == update[e])

                # if 'Parameter_col' is specified, compute input_bools for selecting
                # rows based on the Parameter value in data
                e = 'Parameter'
                if 'Parameter_col' in update:
                    column[e]      = update[e + '_col']
                    input_bools[e] = (data[data_entry][column[e]] == input_param)
                else:
                    input_bools[e]  = [True for _ in range(len(data[data_entry]))]
                # get input data - select rows which comply with the technology,
                # fuel and parameter selection criteria for all years
                input_bool  = [btech & bfuel & bparam for btech,
                               bfuel, bparam in zip(input_bools['TECHNOLOGY'],
                                                input_bools['FUEL'],
                                                input_bools['Parameter'])]
                if sum(input_bool) > 1:
                    for idx in [idx for idx in range(len(input_bool)) if input_bool[idx] is True]:
                        sumnan = sum(np.isnan(x) for x in data[data_entry].loc[idx, years].values)
                        lendata = len(data[data_entry].loc[idx, years].values)
                        if sumnan == lendata:
                            input_bool[idx] = False

                if sum(input_bool) == 0:
                    no_data.append(update)
                elif sum(input_bool) > 1:
                    too_much_data.append(update)
                else:
                    input_data = data[data_entry].loc[input_bool, years].values
                    # update parameter sheet
                    # check if there are conditions for whether to update the df_parameters
                    # right now, only applies to checking if any resid capacity for the tech
                    bool_update = condition_update(param, input_data)
                    if bool_update:
                        output_bool = [btech & bfuel for btech,
                                                         bfuel in zip(output_bools['TECHNOLOGY'],
                                                                      output_bools['FUEL'])]
                        df_parameters.loc[(df_parameters.Parameter == param) &
                                          output_bool, years_str] = input_data
                if len(no_data) > 0:
                    print(f"{country}: no data for {no_data}, too much data for {too_much_data}")

            #%% update parameters where data is available for one year

            # year instance only, to be pasted across all years

            # column   : the column in data to be selected
            # type_col : the name of the column in df_parameters in which we must
            #            check values correspond to the data column

            # reset rules for updating
            updates = []

            # for capacity factor
            # only include offshore wind capfac if it is possible
            techs = [
                'PWRSOL001','PWRSOL002','PWRSOL001S',
                'PWRHYD001', 'PWRHYD002', 'PWRHYD003','PWRHYD004',
                'PWRWND001','PWRWND001S'
            ]
            if offshore_wind_possible:
                techs += ['PWRWND002']
            for tech in techs:
                updates += [{'Parameter'   : 'CapacityFactor',
                           'column'   : tech,
                           'type_col' : 'TECHNOLOGY'}]

            # specified demand profile
            for fuel in fuels_spec_dem: # this list was computed before
                updates += [{'Parameter'   : 'SpecifiedDemandProfile',
                           'column'   : fuel,
                           'type_col' : 'FUEL'}]
            no_data = {}
            for update in updates:
                param    = update['Parameter']
                column   = update['column']
                type_col = update['type_col']

                # select the relevant column in input data, replicate for each year,
                # swap dimensions of the axes
                if column in data[param].keys():
                    input_data = np.swapaxes([data[param][column].values] * len(years_str),
                                             0, 1)
                    # add input_data to df_parameters
                    df_parameters.loc[(df_parameters.Parameter == param) &
                                      (df_parameters[type_col] == column),
                                      years_str] = input_data
                else:
                    if param in no_data:
                        no_data[param].append(column)
                    else:
                        no_data[param] = [column]

                if len(list(no_data.keys())) > 0:
                    print(f"for {country} no data for {no_data}")

            #%% ##### update parameter values which don't fit above patterns

            ###### update total annual max capacity investment
            # add only those which have no residual capacity installed
            # (ensure pwr_techs_rescap does NOT include pwrtrnimp and pwrtrnexp)
            pwr_techs_maxcapinv = []
            for tech in pwr_techs_rescap: # pwr_techs_rescap previously defined
                # get resid cap for each power technology
                res_caps = data['ResidualCapacity'].loc[(data['ResidualCapacity']['Group'] == tech) &
                                           (data['ResidualCapacity']['Power Plant']
                                            =='ResidualCapacity'),
                                                    years].values
                # find sum of resid cap for each power technology
                res_caps_sum = np.nansum(res_caps)
                # if none installed, sum = 0
                if res_caps_sum <= 0:
                    pwr_techs_maxcapinv.append(tech)

            parameter = 'TotalAnnualMaxCapacityInvestment'
            for tech in pwr_techs_maxcapinv:
                # update parameter sheet with 0 for any with no historical investment
                df_parameters.loc[(df_parameters.Parameter == parameter) &
                                   (df_parameters.TECHNOLOGY == tech),
                                      years_past] = np.zeros(len(years_past))

            #%%#### update total tech model period activity upper limit
            param = 'TotalTechnologyModelPeriodActivityUpperLimit'
            model_period_techs = ['MINOIL', 'MINNGS', 'MINCOA']

            for tech in model_period_techs:
                # get the data
                input_data = data['Activity'].loc[(data['Activity']['Technology Code']
                                               == tech) &
                                              (data['Activity']['Constraint']
                                               == param),
                                   'Model period limit where applicable'].values
                # update parameter sheet --- (typo in 'Time indipendent variables')
                df_parameters.loc[(df_parameters.Parameter == param) &
                                   (df_parameters.TECHNOLOGY == tech),
                              'Time indipendent variables'] = input_data[0]

            print(f'parameters are computed for {country}, t = {time.time() - t_start} secs')

            #%%############################################################################
            #%%#########################  Save output  ##########################
            # save copy of parameters sheet as .csv file.
            # if issues with saving macros in .xslm file, this sheet can be copied
            # into Base SAND excel file in single manual step for each country
            df_parameters.to_csv(PATH_SAND_OUT +
                                 country + ' Base SAND parameters.csv',
                                 index = False)
            print(f'parameters are saved for {country}, t = {time.time() - t_start} secs')

    except Exception as ex:
        print(f'ERROR for {country}: {ex}')
        print(traceback.format_exc())
